@@ -20,32 +20,78 @@
  * Microline Widget Library and Plot Widgets library.
  */
 
-#include <Xm/Xm.h>
-#if XmVersion >= 1002
-#include <Xm/PrimitiveP.h>
-#endif
-#include <Xm/Form.h>
-#include <Xm/PushB.h>
-#include <Xm/DrawnB.h>
-#include <Xm/Label.h>
-#include <Xm/Scale.h>
-#include <Xm/TextF.h>
-#include <Xm/FileSB.h>
-
-#include <XmL/Folder.h>
-#include <XmL/Tree.h>
-#include <XmL/Progress.h>
-
-#include <Xm/DrawnB.h>
-
-#include <X11/Intrinsic.h>
-#include "XY.h"
-
-#define _WITH_X
+#include <sys/stat.h>
+#include <unistd.h>
+#include <sys/types.h>
+#include <signal.h>
 
 #include "xcis.h"
+#include "replay.h"
 
 char *printpath = "./";
+static char tmpfilename[16];
+
+static FILE *tmp_file;
+
+extern sigset_t sigblockset;
+
+void folder_activate_cb(Widget w, XtPointer client_data, XtPointer cbsp)
+{
+	XmLFolderCallbackStruct *cbs = (XmLFolderCallbackStruct *) cbsp;
+
+	active_folder = cbs->pos;
+}
+
+void proclist_expand_cb(Widget w, XtPointer client_data, XtPointer cbsp)
+{
+        XmLGridCallbackStruct *cbs = (XmLGridCallbackStruct *) cbsp;
+	struct host_info *h;
+	struct cis_procinfo *ptab = NULL;
+	XmLGridRow rowp;
+	int len;
+
+	if (mode != ONLINE)
+		return;
+
+	rowp = XmLGridGetRow (Proclist, XmCONTENT, cbs->row);
+	XtVaGetValues(Proclist,
+		      XmNrowPtr, rowp,
+		      XmNrowUserData, &h,
+		      NULL);
+
+	sigprocmask(SIG_BLOCK, &sigblockset, NULL);
+
+	ptab = NULL;
+	len = 0;
+	if (h->cis && h->cis->handle) {
+		len = cisProclist (h->cis->handle, &ptab, h->hostinfo->addr, h->uid, 1);
+		if (len == -1) {
+			monitor_error (h, len);
+			len = 0;
+		}
+	}
+	proctab_update (h, ptab, len);
+	proclist_update (h);
+
+	h->flag |= PROC_EXPANDED;
+	
+	sigprocmask(SIG_UNBLOCK, &sigblockset, NULL);
+}
+
+void proclist_collapse_cb(Widget w, XtPointer client_data, XtPointer cbsp)
+{
+	XmLGridCallbackStruct *cbs = (XmLGridCallbackStruct *) cbsp;
+	struct host_info *h;
+	XmLGridRow rowp;
+
+	rowp = XmLGridGetRow (Proclist, XmCONTENT, cbs->row);
+	XtVaGetValues(Proclist,
+		      XmNrowPtr, rowp,
+		      XmNrowUserData, &h,
+		      NULL);
+
+	h->flag &= ~PROC_EXPANDED;
+}
 
 void proclist_cell_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
@@ -54,7 +100,7 @@ void proclist_cell_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
         struct host_info *h;
         
         if (cbs->rowType != XmHEADING)
-                return;
+		return;
 
         if (cbs->columnType != XmCONTENT)
                 return;
@@ -75,6 +121,9 @@ void proclist_cell_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
         case PROC_COL_VM:
                 proclist_sort_type = PROC_SORT_VM;
                 break;
+        case PROC_COL_MAJFLT:
+                proclist_sort_type = PROC_SORT_MAJFLT;
+                break;
         }
 
         XtVaSetValues(Proclist, XmNlayoutFrozen, True, NULL);
@@ -84,11 +133,67 @@ void proclist_cell_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
         XtVaSetValues(Proclist, XmNlayoutFrozen, False, NULL);
 }
 
+void linklist_expand_cb(Widget w, XtPointer client_data, XtPointer cbsp)
+{
+        XmLGridCallbackStruct *cbs = (XmLGridCallbackStruct *) cbsp;
+	struct host_info *h;
+        struct cis_sockinfo *stab = NULL;
+	XmLGridRow rowp;
+	int len;
+
+	if (mode != ONLINE)
+		return;
+
+	rowp = XmLGridGetRow (Linklist, XmCONTENT, cbs->row);
+	XtVaGetValues(Linklist,
+		      XmNrowPtr, rowp,
+		      XmNrowUserData, &h,
+		      NULL);
+
+	sigprocmask(SIG_BLOCK, &sigblockset, NULL);
+
+	stab = NULL;
+	len = 0;
+	if (h->cis && h->cis->handle) {
+		len = cisSocklist (h->cis->handle, &stab, h->hostinfo->addr, h->uid, h->interval);
+		if (len < 0) {
+			monitor_error (h, len);
+			len = 0;
+		}
+	}
+	socktab_update (h, stab, len);
+	socklist_update (h);
+
+	linklist_update();
+
+	h->flag |= LINK_EXPANDED;
+	
+	sigprocmask(SIG_UNBLOCK, &sigblockset, NULL);
+}
+
+void linklist_collapse_cb(Widget w, XtPointer client_data, XtPointer cbsp)
+{
+	XmLGridCallbackStruct *cbs = (XmLGridCallbackStruct *) cbsp;
+	struct host_info *h;
+	XmLGridRow rowp;
+
+	rowp = XmLGridGetRow (Linklist, XmCONTENT, cbs->row);
+	XtVaGetValues(Linklist,
+		      XmNrowPtr, rowp,
+		      XmNrowUserData, &h,
+		      NULL);
+
+	h->flag &= ~LINK_EXPANDED;
+}
+
+
 void linklist_cell_activate_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
         XmLGridCallbackStruct *cbs = (XmLGridCallbackStruct *) cbsp;
         struct host_info *h;
         struct cis_procinfo *process;
+	struct cis_procinfo *ptab = NULL;
+        int len;
         struct link_info *link;
         XmLGridRow rowp;
         unsigned int pid;
@@ -114,9 +219,11 @@ void linklist_cell_activate_cb(Widget w, XtPointer client_data, XtPointer cbsp)
                 return;
         
         switch (cbs->column) {
-        case LINK_COL_PID:
-                for (h = hostlist; h && h->link_offset + h->link_number + 1 < cbs->row; h = h->next)
-			;
+	case LINK_COL_PID:
+		for_each_selected_host(h)
+			if (h->link_offset + h->link_number + 1 >= cbs->row)
+				break;
+
                 pid = link->src_pid;
                 break;
                 
@@ -127,8 +234,21 @@ void linklist_cell_activate_cb(Widget w, XtPointer client_data, XtPointer cbsp)
                         return;
                 pid = link->dst_pid;
                 break;
-        }
+	}
 
+	if (mode == ONLINE) {
+		ptab = NULL;
+		len = 0;
+		if (h->cis && h->cis->handle) {
+			len = cisProclist (h->cis->handle, &ptab, h->hostinfo->addr, h->uid, 1);
+			if (len == -1) {
+				monitor_error (h, len);
+				len = 0;
+			}
+		}
+		proctab_update (h, ptab, len);
+		proclist_update (h);
+	}
         XtVaSetValues(Proclist,
                       XmNrow, h->proc_offset,
                       XmNrowIsExpanded, True,
@@ -136,13 +256,15 @@ void linklist_cell_activate_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 
         for (row = h->proc_offset+1; row < h->proc_offset + h->proc_number + 1; row++) {
                 rowp = XmLGridGetRow (Proclist, XmCONTENT, row);
-                XtVaGetValues(Proclist,
-                              XmNrowPtr, rowp,
+		XtVaGetValues(Proclist,
+			      XmNrowPtr, rowp,
                               XmNrowUserData, &process,
-                              NULL);
+			      NULL);
 
                 if (process->pid == pid) {
-                        XmLFolderSetActiveTab (Folder, PROCLIST, FALSE);
+			XmLFolderSetActiveTab (Folder, PROCLIST, FALSE);
+			active_folder = PROCLIST;
+			h->flag |= PROC_EXPANDED;
                         XmLGridSelectRow (Proclist, row, 0);
                         XmLGridSetFocus (Proclist, row, 0);
                 }
@@ -160,15 +282,19 @@ void record_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
         time_t ct;
         struct tm *tmp;
         char buff[MAXHOSTNAMELEN+32], *newname;
-        struct host_info *h;
-        int len;
-        
+	struct host_info *h;
+	static struct cis_addr addrtab[256];
+	int i, len;
+	struct timeval t;
+
         if (mode == RECORD) {
-                XtVaSetValues (w, XmNshadowType, XmSHADOW_OUT, NULL);
+
+		XtVaSetValues (w, XmNshadowType, XmSHADOW_OUT, NULL);
                 XtVaSetValues (Buttons[REP], XmNsensitive, True, NULL);
                 mode = ONLINE;
                 fclose (record.file);
-        } else {
+
+	} else {
 
                 ct = time (NULL);
                 tmp = localtime (&ct);
@@ -187,15 +313,31 @@ void record_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
                 if (!record.file)
                         return;
 
-                for (h = hostlist; h; h = h->next) {
-                        h->hostinfo_saved_bytes = record.density;
-                        h->proc_saved_bytes     = record.density;
-                        h->sock_saved_bytes     = record.density;
-                        h->dev_saved_bytes      = record.density;
-                }
+		for (i = 0, h = hostlist; h && i < 256;  h = h->next)
+			if (h->flag) {
+				addrtab[i].server_addr = h->hostinfo->caddr;
+				addrtab[i].node_addr   = h->hostinfo->addr;
+				i++;
+			}
+                if (i == 256)
+			printf("reached maximal number of hosts in record file (256)");
+
+		gettimeofday(&t, NULL);
+		if (cisSaveHeader(record.file, &t, i, addrtab, interval) == -1) {
+			printf("error while saving record file header");
+			return;
+		}
+
+		for_each_selected_host(h) {
+			cisSaveHostID (record.file, h->hostinfo);
+			cisSaveHostinfo (record.file, h->hostinfo);
+
+			cisSaveProclist (record.file, h->proctab, h->nproc);
+			cisSaveSocklist (record.file, h->socktab, h->nsock);
+			cisSaveNdevlist (record.file, h->devtab,  h->ndev);
+		}
+		cisSaveBreak (record.file, 0);
 
-                cisSaveHeader (record.file);
-                
                 XtVaSetValues (w, XmNshadowType, XmSHADOW_IN, NULL);
                 XtVaSetValues (Buttons[REP], XmNsensitive, False, NULL);
                 mode = RECORD;
@@ -215,8 +357,9 @@ void replay_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 			XtUnmanageChild(*button);
 
                 fclose (record.file);
-                free (record.table);
-
+		fclose (tmp_file);
+                unlink (tmpfilename);
+		
                 clear_data();
 
                 mode = ONLINE;
@@ -313,9 +456,10 @@ void file_selection_cb(Widget w, XtPointer client_data, XtPointer cbsp)
         static struct itimerval it = {{0,0}, {0,0}};
         XmFileSelectionBoxCallbackStruct *cbs = cbsp;
         char *filename;
-        struct cis_header header;
-	struct record_info *r;
 	Widget *button;
+	struct stat fst;
+	unsigned long length, offset;
+	int i, value, end;
         
         if (cbs->reason == XmCR_OK) {
 
@@ -323,39 +467,79 @@ void file_selection_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 
                 XtUnmanageChild (w);
 
+		setitimer (ITIMER_REAL, &it, NULL);
+
                 record.file = fopen (filename, "r");
                 if (!record.file) {
                         printf ("cannot open file %s\n", filename);
                         goto cancel;
-                }
+		}
+		if (fstat (fileno(record.file), &fst) == -1)
+			goto cancel;
 
-                if (cisReadHeader (record.file, &header) < 0 ||
-                    header.version != CIS_RECORD_VERSION) {
+		record.length = length = fst.st_size;
+
+		if (cisReadHeader (record.file, &record.header) < 0 ||
+		    record.header->version != CIS_RECORD_VERSION) {
                         fclose (record.file);
                         goto cancel;
-                }
+		}
 
-                XtManageChild (Progress_bar);
-                record.table_length = cisParseFile (record.file, &record.table, progress_cb);
-                XtUnmanageChild (Progress_bar);
-                if (record.table_length <= 0) {
-                        fclose (record.file);
-                        goto cancel;
-                }
+		sprintf(tmpfilename, "/tmp/xcis.%d", getpid());
+		tmp_file = fopen (tmpfilename, "w+");
+		if (!tmp_file) {
+			printf ("cannot open temporary file %s\n", tmpfilename);
+			goto cancel;
+		}
 
-                setitimer (ITIMER_REAL, &it, NULL);
+		XtManageChild (Progress_bar);
 
-                record.position = record.table;
-                record.last     = record.table + record.table_length-1;
-                
-                for (r = record.table; r <= record.last; r++)
-                        r->host_ptr = find_host (r->host);
+		clear_data();
+		parse_full_info(record.file);
 
-                old_time = current_time = record.table->time;
+		record.start_time = old_time = current_time = record.header->time;
+		i = time_scale;
+		record.start = ftell(record.file);
 
-                clear_data();
-                parse_next_interval();
+		end = FALSE;
 
+		for(;;) {
+
+			offset = ftell(record.file);
+			value = 100 - ((length - offset)*100)/length;
+			XtVaSetValues (Progress_bar, XmNvalue, value, NULL);
+
+			if (i == time_scale || end) {
+
+				record.end = ftell(tmp_file);
+				cisSaveTime(tmp_file, current_time);
+				save_full_info(tmp_file);
+				cisSaveBreak(tmp_file, offset);
+				save_curves(tmp_file);
+				cisSaveBreak(tmp_file, ftell(tmp_file) - record.end);
+
+				i = 0;
+			}
+			if (end)
+				break;
+
+			end = !parse_next_interval();
+
+			i++;
+		}
+
+		XtUnmanageChild (Progress_bar);
+
+		old_time = current_time = record.start_time;
+
+		clear_data();
+		rewind(tmp_file);
+		parse_full_info(tmp_file);
+		fseek(record.file, record.offset, SEEK_SET);
+                parse_curves (tmp_file);
+
+		record.counter = 0;
+		
                 update_time ();
 
 		for (button = Buttons+BEGIN; button <= Buttons+END; button++)
@@ -366,6 +550,7 @@ void file_selection_cb(Widget w, XtPointer client_data, XtPointer cbsp)
                 mode = REPLAY;
 
                 XmUpdateDisplay (Xcis);
+		XtNoticeSignal (sigid);
 
                 return;
         }
@@ -380,48 +565,116 @@ XtIntervalId timer;
 
 void timer_cb (XtPointer button_cbp, XtIntervalId *id)
 {
-        ((XtCallbackProc) button_cbp) (NULL,NULL,NULL);
+	if (!timer)
+		return;
+	((XtCallbackProc) button_cbp) (NULL,NULL,NULL);
 }
 
 void disarm_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
-        XtRemoveTimeOut(timer);
+	if (!timer)
+		return;
+	XtRemoveTimeOut(timer);
+	timer = 0;
 }
 
 void begin_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
-        rewind_record_file (record.table->time);
+	rewind(tmp_file);
+	parse_full_info(tmp_file);
+	fseek(record.file, record.offset, SEEK_SET);
+	parse_curves (tmp_file);
+	record.counter = 0;
+
+	update_time ();
+	XtNoticeSignal (sigid);
 }
 
 void rew_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
-        struct timeval tv = current_time;
+	if (record.offset == record.start)
+		return;
 
-        tvalsub (&tv, record.jump);
+	if (!record.counter)
+		fseek(tmp_file, -(get_offset(tmp_file) + CIS_RECHDRLEN + sizeof(int)), SEEK_CUR);
+	fseek(tmp_file, -(get_offset(tmp_file) + CIS_RECHDRLEN + sizeof(int)), SEEK_CUR);
 
-        rewind_record_file (tv);
-        timer = XtAppAddTimeOut (app_context, DELAY, timer_cb, rew_button_cb);
+	parse_full_info(tmp_file);
+	fseek(record.file, record.offset, SEEK_SET);
+	old_time = current_time;
+	parse_curves (tmp_file);
+	record.counter = 0;
+
+	timer = XtAppAddTimeOut (app_context, DELAY, timer_cb, rew_button_cb);
+
+	update_time ();
+	XtNoticeSignal (sigid);
 }
 
 void step_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
-        parse_next_interval();
-        timer = XtAppAddTimeOut (app_context, DELAY, timer_cb, step_button_cb);
+	struct record_info r;
+	
+	if (record.offset == record.length)
+		return;
+
+	if (record.counter == time_scale) {
+		/*
+		 * Go to next full info
+		 */
+		while (cisRecordInfo(tmp_file, &r) == 0 && r.type != CIS_REC_BREAK)
+			fseek(tmp_file, r.length + CIS_RECHDRLEN, SEEK_CUR);
+
+		fseek(tmp_file, r.length + CIS_RECHDRLEN, SEEK_CUR);
+
+		while (cisRecordInfo(tmp_file, &r) == 0 && r.type != CIS_REC_BREAK)
+			fseek(tmp_file, r.length + CIS_RECHDRLEN, SEEK_CUR);
+
+		fseek(tmp_file, r.length + CIS_RECHDRLEN, SEEK_CUR);
+	}
+
+	parse_next_interval();
+        record.counter++;
+
+	timer = XtAppAddTimeOut (app_context, DELAY, timer_cb, step_button_cb);
+
+	update_time ();
+	XtNoticeSignal (sigid);
 }
 
 void fwd_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
-        struct timeval tv = current_time;
+	if (record.offset == record.length)
+		return;
 
-        tvaladd (&tv, record.jump);
+	parse_full_info(tmp_file);
 
-        rewind_record_file (tv);
-        timer = XtAppAddTimeOut (app_context, DELAY, timer_cb, fwd_button_cb);
+	fseek(record.file, record.offset, SEEK_SET);
+	old_time = current_time;
+	parse_curves (tmp_file);
+	record.counter = 0;
+
+	timer = XtAppAddTimeOut (app_context, DELAY, timer_cb, fwd_button_cb);
+
+	update_time ();
+	XtNoticeSignal (sigid);
 }
 
 void end_button_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 {
-        rewind_record_file (record.last->time);
+	if (record.offset == record.length)
+		return;
+
+	fseek(tmp_file, record.end, SEEK_SET);
+	record.counter = 0;
+
+	parse_full_info(tmp_file);
+	fseek(record.file, record.offset, SEEK_SET);
+	old_time = current_time;
+	parse_curves (tmp_file);
+
+	update_time ();
+	XtNoticeSignal (sigid);
 }
 
 void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
@@ -432,7 +685,8 @@ void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
         Pixmap mask;
         struct host_info *host, *h;
         XmString s;
-        XYCurve *c;
+	XYCurve *c;
+	XYPoint *points;
         int old_type;
 
         if (cbs->reason != XmCR_SELECT_CELL)
@@ -459,6 +713,9 @@ void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
                 hostlist_sort ();
                 return;
         }
+
+	if (mode != ONLINE)
+		return;
         
         row = XmLGridGetRow(w, cbs->rowType, cbs->row);
         column = XmLGridGetColumn(w, XmCONTENT, 0);
@@ -471,9 +728,16 @@ void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
                       NULL);
         
 	if (mask == pixmasks[UNCHECK]) {
+		
+		points = calloc (time_scale, sizeof (XYPoint));
+		if (!points) {
+			printf("not enough memory\n");
+			return;
+		}
+
                 mask = pixmasks[CHECK];
 
-                s = XmStringCreateSimple(host->name);
+		s = XmStringCreateSimple(host->hostinfo->name);
                 XmLTreeAddRow(Proclist, 0, True, False, host->proc_offset,
                               XmUNSPECIFIED_PIXMAP, XmUNSPECIFIED_PIXMAP, s);
                 XmLTreeAddRow(Linklist, 0, True, False, host->link_offset,
@@ -481,7 +745,11 @@ void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 //                XmLTreeAddRow(Devlist, 0, True, True, host->dev_offset,
 //                              XmUNSPECIFIED_PIXMAP, XmUNSPECIFIED_PIXMAP, s);
 
-                XmStringFree (s);
+		XmStringFree (s);
+		XtVaSetValues (Proclist, XmNrow, host->proc_offset,
+			       XmNrowUserData, host, NULL);
+		XtVaSetValues (Linklist, XmNrow, host->link_offset,
+			       XmNrowUserData, host, NULL);
 
 		for (h = host->next; h; h = h->next) {
 			h->proc_offset++;
@@ -503,8 +771,9 @@ void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 		} else
 			c = CPU_curves+nCPU_curves;
 		
-		c->name   = XmStringCreateSimple(host->name);
-                c->points = NULL;
+		c->name   = XmStringCreateSimple(host->hostinfo->name);
+
+                c->points = points;
                 c->nPoints = 0;
                 
 		host->CPU_curve = c;
@@ -555,3 +824,30 @@ void host_grid_select_cb(Widget w, XtPointer client_data, XtPointer cbsp)
 
 	XYSetCurves (CPU_plot, CPU_curves, nCPU_curves, XY_NO_RESCALE, True);
 }
+
+void allarm_cb (XtPointer cl_data, XtSignalId *id)
+{
+        struct host_info *h;
+        int update_links = FALSE;
+
+	for (h = hostlist; h; h = h->next) {
+		if (h->hostinfo_updated)
+                        hostlist_update (h);
+                if (h->proc_updated)
+                        proclist_update (h);
+                if (h->sock_updated) {
+                        socklist_update (h);
+                        update_links = TRUE;
+                }
+                if (h->dev_updated)
+                        devlist_update (h);
+        }
+        if (update_links)
+                linklist_update();
+
+        if (hostlist_sort_type != HOST_SORT_NONE)
+                hostlist_sort ();
+
+        update_curves_time (1);
+        update_time ();
+}
