@@ -20,30 +20,9 @@
  * Microline Widget Library and Plot Widgets library.
  */
 
-#include <Xm/Xm.h>
-#if XmVersion >= 1002
-#include <Xm/PrimitiveP.h>
-#endif
-#include <Xm/Form.h>
-#include <Xm/PushB.h>
-#include <Xm/DrawnB.h>
-#include <Xm/Label.h>
-#include <Xm/Scale.h>
-#include <Xm/TextF.h>
-#include <Xm/FileSB.h>
-
-#include <XmL/Folder.h>
-#include <XmL/Tree.h>
-#include <XmL/Progress.h>
-
-#include <Xm/DrawnB.h>
-
-#include <X11/Intrinsic.h>
-#include "XY.h"
-
-#define _WITH_X
-
 #include "xcis.h"
+#include "replay.h"
+#include "cb.h"
 
 char title[] ="Xcis";
 char *folder_title[NFOLDERS] = {"Hosts", "Processes", "Links", "Net devices", "CPU", "Net usage"};
@@ -165,7 +144,7 @@ void create_window(int argc, char *argv[])
         XtSetArg(args[n], XmNmarginWidth, 0); n++;
         XtSetArg(args[n], XmNshadowThickness, 0); n++;
         XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
-        XtSetArg(args[n], XmNbottomOffset, 0); n++;
+//        XtSetArg(args[n], XmNbottomOffset, 0); n++;
         XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
         Time_field = XmCreateLabel (top_form, "time_field", args, n);
         XtManageChild(Time_field);
@@ -174,107 +153,14 @@ void create_window(int argc, char *argv[])
 	update_time ();
         old_time = current_time;
         
-        n = 0;
-	s = XmStringCreateSimple ("PRINT");
-        XtSetArg(args[n], XmNlabelString, s); n++;
-        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
-        XtSetArg(args[n], XmNbottomOffset, 0); n++;
-        XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
-        XtSetArg(args[n], XmNmarginHeight, 1); n++;
-        Buttons[PRINT]  = XmCreatePushButton  (top_form, "print_button", args, n);
-        XtAddCallback(Buttons[PRINT], XmNarmCallback, print_button_cb, NULL);
-        XmStringFree (s);
-
-        n = 0;
-        XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
-        XtSetArg(args[n], XmNtopWidget, Buttons[PRINT]); n++;
-        XtSetArg(args[n], XmNtopOffset, 0); n++;
-        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
-        XtSetArg(args[n], XmNbottomOffset, 0); n++;
-        XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
-        XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
-
-        s = XmStringCreateSimple ("RECORD");
-        XtSetArg(args[n], XmNlabelString, s); n++;
-        XtSetArg(args[n], XmNleftWidget, Buttons[PRINT]); n++;
-        Buttons[REC] = XmCreateDrawnButton (top_form, "Buttons[REC]", args, n);
-        XmStringFree (s);
-
-        s = XmStringCreateSimple ("REPLAY");
-        XtSetArg(args[n-1], XmNlabelString, s);
-        XtSetArg(args[n-2], XmNleftWidget, Buttons[REC]);
-        Buttons[REP] = XmCreateDrawnButton (top_form, "replay_button", args, n);
-        XmStringFree (s);
-
-        n -= 2;
-        XtSetArg(args[n], XmNwidth, 20); n++;
-        Buttons[BEGIN] = XmCreateDrawnButton (top_form, "begin_button", args, n);
-        XtSetArg(args[n], XmNleftOffset, 0); n++;
-        Buttons[REW]   = XmCreateDrawnButton (top_form, "rew_button", args, n);
-        Buttons[STEP]  = XmCreateDrawnButton (top_form, "step_button", args, n);
-        Buttons[FWD]   = XmCreateDrawnButton (top_form, "fwd_button", args, n);
-        Buttons[END]   = XmCreateDrawnButton (top_form, "end_button", args, n);
-
-        XtAddCallback(Buttons[REC], XmNarmCallback, record_button_cb, NULL);
-        XtAddCallback(Buttons[REP], XmNarmCallback, replay_button_cb, NULL);
-
-        XtVaSetValues(Buttons[BEGIN], XmNleftWidget, Buttons[REP], NULL);
-        XmLDrawnButtonSetType(Buttons[BEGIN], XmDRAWNB_ARROWLINE, XmDRAWNB_LEFT);
-        XtAddCallback(Buttons[BEGIN], XmNarmCallback, begin_button_cb, NULL);
-        
-        XtVaSetValues(Buttons[REW], XmNleftWidget, Buttons[BEGIN], NULL);
-        XmLDrawnButtonSetType(Buttons[REW], XmDRAWNB_DOUBLEARROW, XmDRAWNB_LEFT);
-        XtAddCallback(Buttons[REW], XmNarmCallback, rew_button_cb, NULL);
-        XtAddCallback(Buttons[REW], XmNdisarmCallback, disarm_cb, NULL);
-
-        XtVaSetValues(Buttons[STEP], XmNleftWidget, Buttons[REW], NULL);
-        XmLDrawnButtonSetType(Buttons[STEP], XmDRAWNB_ARROW, XmDRAWNB_RIGHT);
-        XtAddCallback(Buttons[STEP], XmNarmCallback, step_button_cb, NULL);
-        XtAddCallback(Buttons[STEP], XmNdisarmCallback, disarm_cb, NULL);
-
-        XtVaSetValues(Buttons[FWD], XmNleftWidget, Buttons[STEP], NULL);
-        XmLDrawnButtonSetType(Buttons[FWD], XmDRAWNB_DOUBLEARROW, XmDRAWNB_RIGHT);
-        XtAddCallback(Buttons[FWD], XmNarmCallback, fwd_button_cb, NULL);
-        XtAddCallback(Buttons[FWD], XmNdisarmCallback, disarm_cb, NULL);
-
-        XtVaSetValues(Buttons[END], XmNleftWidget, Buttons[FWD], NULL);
-        XmLDrawnButtonSetType(Buttons[END], XmDRAWNB_ARROWLINE, XmDRAWNB_RIGHT);
-        XtAddCallback(Buttons[END], XmNarmCallback, end_button_cb, NULL);
-
-        XtManageChild(Buttons[PRINT]);
-        XtManageChild(Buttons[REC]);
-        XtManageChild(Buttons[REP]);
-
-        n = 0;
-        XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
-        XtSetArg(args[n], XmNtopWidget, Buttons[PRINT]); n++;
-        XtSetArg(args[n], XmNtopOffset, 0); n++;
-        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
-        XtSetArg(args[n], XmNbottomOffset, 0); n++;
-        XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
-        XtSetArg(args[n], XmNleftWidget, Buttons[REP]); n++;
-        XtSetArg(args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
-        XtSetArg(args[n], XmNrightWidget, Time_field); n++;
-        XtSetArg(args[n], XmNcompleteValue, 100); n++;
-        XtSetArg(args[n], XmNmeterStyle, XmMETER_BOXES); n++;
-        XtSetArg(args[n], XmNnumBoxes, 20); n++;
-	XtSetArg(args[n], XmNshadowThickness, 1); n++;
-
-        Progress_bar = XmLCreateProgress (top_form, "progress_bar", args, n);
-
-        XtVaSetValues (Progress_bar,
-                       XmNvalue, 50,
-                       XtVaTypedArg, XmNbackground, XmRString, "white", 6,
-                       XtVaTypedArg, XmNforeground, XmRString, "#000080", 8,
-                       NULL);
-
         /*
          * Create folder.
          */
         n = 0;
         XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
         XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
-        XtSetArg(args[n], XmNbottomWidget, Buttons[PRINT]); n++;
+//        XtSetArg(args[n], XmNbottomWidget, Buttons[PRINT]); n++;
+        XtSetArg(args[n], XmNbottomWidget, Time_field); n++;
         XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
         XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
         
@@ -287,6 +173,8 @@ void create_window(int argc, char *argv[])
                 XmStringFree(s);
         }
 
+	XtAddCallback(Folder, XmNactivateCallback, folder_activate_cb, NULL);
+
         /*
          * Create tree widgets.
          */
@@ -310,7 +198,7 @@ void create_window(int argc, char *argv[])
         XtSetArg(args[n], XmNcolumns, HOST_NCOLS); n++;
         XtSetArg(args[n], XmNallowColumnResize, True); n++;
         XtSetArg(args[n], XmNcellDefaults, True); n++;
-        XtSetArg(args[n], XmNsimpleWidths, "2c 15c 6c 6c 8c 8c 8c 8c 5c 7c 7c 7c 10c"); n++;
+        XtSetArg(args[n], XmNsimpleWidths, "2c 15c 6c 6c 8c 8c 8c 8c 5c 7c 7c 7c 7c 10c"); n++;
         XtSetArg(args[n], XmNselectionPolicy, XmSELECT_NONE); n++;
         XtSetArg(args[n], XmNcellLeftBorderType, XmBORDER_NONE); n++;
         XtSetArg(args[n], XmNcellRightBorderType, XmBORDER_NONE); n++;
@@ -341,7 +229,7 @@ void create_window(int argc, char *argv[])
                        XmNcellColumnSpan, 1,
                        NULL);
 
-        XtVaSetValues (Hostlist,
+	XtVaSetValues (Hostlist,
                        XmNrowType, XmHEADING,
                        XmNrow, 0,
                        XmNcolumn, 6,
@@ -367,17 +255,24 @@ void create_window(int argc, char *argv[])
                        XmNcellRowSpan, 1,
                        NULL);
 
-        XtVaSetValues (Hostlist,
+	XtVaSetValues (Hostlist,
                        XmNrowType, XmHEADING,
                        XmNrow, 0,
                        XmNcolumn, 9,
+                       XmNcellRowSpan, 1,
+                       NULL);
+
+        XtVaSetValues (Hostlist,
+                       XmNrowType, XmHEADING,
+                       XmNrow, 0,
+                       XmNcolumn, 10,
                        XmNcellColumnSpan, 2,
                        NULL);
 
 	XtVaSetValues (Hostlist,
                        XmNrowType, XmHEADING,
                        XmNrow, 0,
-                       XmNcolumn, 12,
+                       XmNcolumn, 13,
                        XmNcellRowSpan, 1,
                        NULL);
 
@@ -397,8 +292,8 @@ void create_window(int argc, char *argv[])
 
         
         XmLGridSetStrings (Hostlist,
-                           "Name||Performance||RAM size in KB||Swap size in KB||Proc|Average load|||Reliability\n|"
-                           "|max.|avail.|total|free|total|free||1min|5min|15min||");
+                           "Name||Performance||RAM size in kB||Swap size in kB||Proc|Ctx_sw|Average load|||Reliability\n|"
+                           "|max.|avail.|total|free|total|free|||1min|5min|15min||");
 
 	XtAddCallback(Hostlist, XmNselectCallback, host_grid_select_cb, NULL);
 
@@ -411,8 +306,8 @@ void create_window(int argc, char *argv[])
         XtSetArg(args[n], XmNcolumns, PROC_NCOLS); n++;
         XtSetArg(args[n], XmNallowColumnResize, True); n++;
         XtSetArg(args[n], XmNcellDefaults, True); n++;
-        XtSetArg(args[n], XmNsimpleHeadings, "Command|PID|User|Pri|RSS|VM|%CPU|Uptime|Usage"); n++;
-        XtSetArg(args[n], XmNsimpleWidths, "15c 7c 7c 4c 7c 7c 7c 7c 7c"); n++;
+        XtSetArg(args[n], XmNsimpleHeadings, "Command|PID|User|Pri|RSS|VM|%CPU|Majflt|Read|Write|Time|Usage"); n++;
+        XtSetArg(args[n], XmNsimpleWidths, "15c 7c 7c 4c 7c 7c 7c 7c 7c 7c 7c 7c"); n++;
         XtSetArg(args[n], XmNcellLeftBorderType, XmBORDER_NONE); n++;
         XtSetArg(args[n], XmNcellRightBorderType, XmBORDER_NONE); n++;
         XtSetArg(args[n], XmNcellTopBorderType, XmBORDER_NONE); n++;
@@ -486,9 +381,13 @@ void create_window(int argc, char *argv[])
                        XmNcolumn, DEV_COL_STATUS,
                        XmNcellAlignment, XmALIGNMENT_CENTER, NULL);
 
+	XtAddCallback(Proclist, XmNexpandCallback, proclist_expand_cb, NULL);
+	XtAddCallback(Proclist, XmNcollapseCallback, proclist_collapse_cb, NULL);
         XtAddCallback(Proclist, XmNselectCallback, proclist_cell_select_cb, NULL);
         XtAddCallback(Proclist, XmNscrollCallback, scroll_cb, NULL);
-        XtAddCallback(Linklist, XmNactivateCallback, linklist_cell_activate_cb, NULL);
+	XtAddCallback(Linklist, XmNexpandCallback, linklist_expand_cb, NULL);
+	XtAddCallback(Linklist, XmNcollapseCallback, linklist_collapse_cb, NULL);
+	XtAddCallback(Linklist, XmNactivateCallback, linklist_cell_activate_cb, NULL);
         
         XtManageChild(Proclist);
         XtManageChild(form[PROCLIST]);
@@ -518,7 +417,7 @@ void create_window(int argc, char *argv[])
 
         XtManageChild(CPU_plot);
         XtManageChild(form[CPUPLOT]);
-        XYSetVisibleRange (CPU_plot,-time_scale , 0, 0, CPU_scale);
+        XYSetVisibleRange (CPU_plot, -(time_scale-1)*interval, 0, 0, CPU_scale);
 
         /*
          * Net plot widget
@@ -538,9 +437,108 @@ void create_window(int argc, char *argv[])
 
         XtManageChild(net_plot);
         XtManageChild(form[NETPLOT]);
-        XYSetVisibleRange (net_plot,-time_scale , 0, 0, net_scale);
+        XYSetVisibleRange (net_plot,-(time_scale-1)*interval , 0, 0, net_scale);
 
         XtManageChild(Folder);
+
+	n = 0;
+	s = XmStringCreateSimple ("PRINT");
+        XtSetArg(args[n], XmNlabelString, s); n++;
+        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
+//        XtSetArg(args[n], XmNbottomOffset, 0); n++;
+        XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
+        XtSetArg(args[n], XmNmarginHeight, 1); n++;
+        Buttons[PRINT]  = XmCreatePushButton  (top_form, "print_button", args, n);
+        XtAddCallback(Buttons[PRINT], XmNarmCallback, print_button_cb, NULL);
+        XmStringFree (s);
+
+        n = 0;
+        XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
+        XtSetArg(args[n], XmNtopWidget, Buttons[PRINT]); n++;
+	XtSetArg(args[n], XmNtopOffset, 0); n++;
+        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
+//        XtSetArg(args[n], XmNbottomOffset, 0); n++;
+        XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
+        XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
+        XtSetArg(args[n], XmNmarginHeight, 1); n++;
+
+        s = XmStringCreateSimple ("RECORD");
+        XtSetArg(args[n], XmNlabelString, s); n++;
+        XtSetArg(args[n], XmNleftWidget, Buttons[PRINT]); n++;
+        Buttons[REC] = XmCreateDrawnButton (top_form, "Buttons[REC]", args, n);
+        XmStringFree (s);
+
+        s = XmStringCreateSimple ("REPLAY");
+        XtSetArg(args[n-1], XmNlabelString, s);
+        XtSetArg(args[n-2], XmNleftWidget, Buttons[REC]);
+        Buttons[REP] = XmCreateDrawnButton (top_form, "replay_button", args, n);
+        XmStringFree (s);
+
+        n -= 3;
+        XtSetArg(args[n], XmNwidth, 20); n++;
+        XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
+        XtSetArg(args[n], XmNtopWidget, Buttons[PRINT]); n++;
+	XtSetArg(args[n], XmNtopOffset, 0); n++;
+        Buttons[BEGIN] = XmCreateDrawnButton (top_form, "begin_button", args, n);
+        XtSetArg(args[n], XmNleftOffset, 0); n++;
+        Buttons[REW]   = XmCreateDrawnButton (top_form, "rew_button", args, n);
+        Buttons[STEP]  = XmCreateDrawnButton (top_form, "step_button", args, n);
+        Buttons[FWD]   = XmCreateDrawnButton (top_form, "fwd_button", args, n);
+        Buttons[END]   = XmCreateDrawnButton (top_form, "end_button", args, n);
+
+        XtAddCallback(Buttons[REC], XmNarmCallback, record_button_cb, NULL);
+        XtAddCallback(Buttons[REP], XmNarmCallback, replay_button_cb, NULL);
+
+        XtVaSetValues(Buttons[BEGIN], XmNleftWidget, Buttons[REP], NULL);
+        XmLDrawnButtonSetType(Buttons[BEGIN], XmDRAWNB_ARROWLINE, XmDRAWNB_LEFT);
+        XtAddCallback(Buttons[BEGIN], XmNarmCallback, begin_button_cb, NULL);
+        
+        XtVaSetValues(Buttons[REW], XmNleftWidget, Buttons[BEGIN], NULL);
+        XmLDrawnButtonSetType(Buttons[REW], XmDRAWNB_DOUBLEARROW, XmDRAWNB_LEFT);
+        XtAddCallback(Buttons[REW], XmNarmCallback, rew_button_cb, NULL);
+        XtAddCallback(Buttons[REW], XmNdisarmCallback, disarm_cb, NULL);
+
+        XtVaSetValues(Buttons[STEP], XmNleftWidget, Buttons[REW], NULL);
+        XmLDrawnButtonSetType(Buttons[STEP], XmDRAWNB_ARROW, XmDRAWNB_RIGHT);
+        XtAddCallback(Buttons[STEP], XmNarmCallback, step_button_cb, NULL);
+        XtAddCallback(Buttons[STEP], XmNdisarmCallback, disarm_cb, NULL);
+
+        XtVaSetValues(Buttons[FWD], XmNleftWidget, Buttons[STEP], NULL);
+        XmLDrawnButtonSetType(Buttons[FWD], XmDRAWNB_DOUBLEARROW, XmDRAWNB_RIGHT);
+        XtAddCallback(Buttons[FWD], XmNarmCallback, fwd_button_cb, NULL);
+        XtAddCallback(Buttons[FWD], XmNdisarmCallback, disarm_cb, NULL);
+
+        XtVaSetValues(Buttons[END], XmNleftWidget, Buttons[FWD], NULL);
+        XmLDrawnButtonSetType(Buttons[END], XmDRAWNB_ARROWLINE, XmDRAWNB_RIGHT);
+        XtAddCallback(Buttons[END], XmNarmCallback, end_button_cb, NULL);
+
+        XtManageChild(Buttons[PRINT]);
+        XtManageChild(Buttons[REC]);
+        XtManageChild(Buttons[REP]);
+
+        n = 0;
+        XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
+        XtSetArg(args[n], XmNtopWidget, Buttons[PRINT]); n++;
+        XtSetArg(args[n], XmNtopOffset, 0); n++;
+        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
+//        XtSetArg(args[n], XmNbottomOffset, 0); n++;
+        XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
+        XtSetArg(args[n], XmNleftWidget, Buttons[REP]); n++;
+        XtSetArg(args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
+        XtSetArg(args[n], XmNrightWidget, Time_field); n++;
+        XtSetArg(args[n], XmNcompleteValue, 100); n++;
+        XtSetArg(args[n], XmNmeterStyle, XmMETER_BOXES); n++;
+        XtSetArg(args[n], XmNnumBoxes, 20); n++;
+	XtSetArg(args[n], XmNshadowThickness, 1); n++;
+
+        Progress_bar = XmLCreateProgress (top_form, "progress_bar", args, n);
+
+        XtVaSetValues (Progress_bar,
+                       XmNvalue, 50,
+                       XtVaTypedArg, XmNbackground, XmRString, "white", 6,
+                       XtVaTypedArg, XmNforeground, XmRString, "#000080", 8,
+                       NULL);
+
         XtManageChild(top_form);
 
         n = 0;
